// ==UserScript==
// @name           SbbItinerary
// @namespace      net.bernhardwagner
// @description    Provides a text-only representation of itineraries from fahrplan.sbb.ch
// @include        http://fahrplan.sbb.ch/bin/query.exe/dn?*
// @include        file:///Users/bwagner/projects/greaseMonkey/sbb/*.html*
// ==/UserScript==
// Version 0.2
// Changes:
//   Version 0.2:
// - removed Legend, Fahrplanauskünfte, Reiseinformationen, Anfrage per SMS, ... Nummer 222, Impressum, etc.
// - if all entries start with same city, it is removed from the station names.
// - abbreviates platz to pl. and strasse to str.
// - abbreviates Tram to Trm
//   Version 0.1:
// - removes ads
// - adds a table with all entries as text, gives it unique id ("iti#), where # is a sequence number
// - the Details buttons (1, 2, 3, 4) in the beginning get added function:
//   They not only jump to the respective itinerary, but also select the respective generated table,
//   thus allowing the immediate copying to clipboard (saving you the selecting of all entries)
//   http://www.oreillynet.com/lpt/a/6257
//
// TODO:
//
// - Selection only works after changing to another application!
//
// - Paste in my own google ads, hehehehehe....
// - Look for TODO/FIXME in entire source code.
// - optionally allow the adding of remarks when displaying something containing "Richtung"
// 		Usually remarks are suppressed in the output (except for Fussweg time).
// 		When displaying Trams or Buses it might be useful to see what direction needs to be taken.
// 		The entries here need to be abbreviated too. (Richtung -> Richtg)
// 		The word Tram, Bus can be entirely omitted, we already know that from "Reise mit" (product)
// - provide language specific abbreviations
// 	- how do I know what language is being displayed?
// - how to provide a download link with the slimmed down itinerary?
// - do I need to provide a namespace to avoid cluttering of the environment?
//   - provide a function that allows to copy the stuff to the clipboard.
//   		- how does tinyurl do it?? probably because it's a "proper" plugin it has more privileges.
//   		- can I reverse engineer a plugin? Disassemble?
//   - BugFirstLineEmpty: there's an empty row at the beginning.
// 	 
//
// TODO: learn how to use xpath from a dom subtree in javascript?
// TODO: local way to compile instead of using http://arantius.com/misc/greasemonkey/script-compiler
// e.g.: http://github.com/ginatrapani/greasemonkey-multi-script-compiler#readme
// TODO: add icon, gui, preferences
//   - whether ads should be removed or not
//   	- which details to remove
//   - highlight color
//   - (personalized?) abbreviations
//   - reacting to language setting
//
//   Intercepting URLs:
//   http://forums.mozillazine.org/viewtopic.php?f=19&t=501254&start=0&st=0&sk=t&sd=a

//GM_log("running sbb fahrplan site");

//----beginning of main --------------------------------------------------------


//GM_log("2 running sbb fahrplan site");


removeAdStuff();
//GM_log("3 removed ads");

provideSelectionSourceCode();

//GM_log("4 done provideSelectionSourceCode");

function parseDetails(itinerary) {

	var itineraryRows = getElementsAtXpath('//*[@class="hac_detail"][@summary]/tbody/tr');
	// problem with this approach: itineraries get flattened out.
	// So we need to recover the structure:
	// - when a tr has th as children: this is a heading row, skip it. 
	// - when a tr has a td as child with colspan >=9, it's the last row: append collected text
	// - when a tr has td as children: this is an itinerary row, join tds

	if(itineraryRows.length == 0){
		GM_log("no itineraryRows found");
		return;
	} else {

	}

	// Try writing a different greasemonkey script that simply checks whether there are any nodes that
	// match my expectation, i.e. a //table[@class = "hac_detail"]
	// Unfortunately that doesn't work. When clicking on a detail button, Greasemonkey is not called!
	// TODO: Find the appropriate way to get called.

	var squeezed = new Array();
	//  iterating over tr elements representing the rows of the itinerary.
	//  Apparently, itineraryRows[i].childNodes[0] is always some empty text.
	//  so we start with itineraryRows[i].childNodes[1].

	var functionMap = initializeFunctionMap();

	for (i = 0; i < itineraryRows.length; ++i){
		
		var itineraryRow = itineraryRows[i];

		var headers = itineraryRow.childNodes[3];
		var currentItinerary;
		if(headers != null) {
			headers = headers.getAttribute( 'headers' );
			if(headers != null) {
				currentItinerary = headers.substring("stops-C1-".length);
//				GM_log("currentItinerary:" + currentItinerary + " itinerary requested: " + itinerary);
				if (currentItinerary != itinerary) {
//					GM_log("currentItinerary:" + currentItinerary + " != " + itinerary +", skipping.");
					continue;
				}
			}
		}

		if(itineraryRow.childNodes[1].getAttribute('colspan') >= 9) { // last line of this itinerary.

			if (currentItinerary == itinerary) {
				appendRow(itineraryRow, squeezed, itinerary);
			}

			squeezed = new Array();
			continue;

		}

		var cells = new Array(itineraryRow.childNodes.length);
		//GM_log("itinerary row: "+i );
		for(var j=0; j < cells.length; ++j){
			var currentChildNode = itineraryRow.childNodes[j];
			if (currentChildNode.firstChild == null) {
				cells[j] = "";
			} else {
				if(currentChildNode.nodeName != "TD") {
					continue;
				}
				var clazz = currentChildNode.getAttribute('class');
				if(clazz == null){
					continue;
				}
				if(clazz == ""){
					continue;
				}
				clazz = normalize(clazz);
				var func = functionMap[clazz];
				cells[j] = func != null ? cells[j] = func(currentChildNode) : 
					cells[j] = "unknown class:" + clazz + currentChildNode.firstChild.nodeValue + " " + j + " ";
			}
			//GM_log("itinerary row: "+i+" colum "+j +" "+cells[j] );
		}
		//GM_log("itinerary row: "+i + " cells:" + cells);
		squeezed.push(cells);
		
	}
//	GM_log("ok, done.");
	//---- inner functions -------------------------------------------------------
	//----------------------------------------------------------------------------
	function normalize(name){
		return name.replace(/ /g, "_").replace(/-/g, "_");
	}

	function initializeFunctionMap() {
		var fussweg = false; // closure variable

		return {
			stop_station_icon:               fMdontCare,
			stop_station_icon_last:          fMdontCare,
			stop_station:                    fMstopStation,
			stop_station_last:               fMstopStation,
			date:                            fMdontCare,
			date_last:                       fMdontCare,
			time_prefix_timeLeft:            fMDefault,
			time_prefix_timeRight:           fMDefault,
			time_prefix_last_timeLeft:       fMDefault,
			time_prefix_last_timeRight:      fMDefault,
			time_last_timeLeft:              fMDefault,
			time_last_timeRight:             fMDefault,
			time_timeLeft:                   fMDefault,
			time_timeRight:                  fMDefault,
			platform:                        fMplatform,
			platform_last:                   fMplatform,
			products_last:                   fMproducts,
			capacity_last:                   fMdontCare,
			remarks_last:                    fMremarks,
			prognosis:                       fMDefault,
			menuLinks_noPrint_last:          fMDefault,
			stop_station_icon_noPrint:       fMDefault,
			stop_station_icon_noPrint_last:  fMDefault,
			noPrint:                         fMDefault,
			last_noPrint:                    fMDefault,
		};

		//------------------------------------------------------------------------------
		function fMdontCare(node) {
			return "";
		}

		//------------------------------------------------------------------------------
		// Reads variable "fussweg" as a closure.
		function fMremarks(node) {

			if(!fussweg){
				return fMdontCare(currentChildNode);
			}

			var imgNodes = node.childNodes;
			var strArr = new Array;
			var j= 0;
			for(var i = 0; i < imgNodes.length ; ++i){
				if(imgNodes[i].nodeName == "IMG"){
					strArr[j++] = imgNodes[i].getAttribute('alt');
				}
			}
			return fMDefault(node) + strArr.join(", ");

		}

		//------------------------------------------------------------------------------
		function fMplatform(node) {
			var plf = node.firstChild.nodeValue;
			if(plf != null && plf.match(/\d/)) {
				plf = "Gleis "+plf+" ";
			}
			return plf;
		}

		//------------------------------------------------------------------------------
		function fMDefault(node) {
			return node.firstChild.nodeValue ;
		}

		//------------------------------------------------------------------------------
		function capacity(node) {
			var img1 = node.childNodes[1].childNodes[1].childNodes[1];
			var img2 = node.childNodes[1].childNodes[3].childNodes[1];
			if (img1 != null) {
				return "Belegung: 1. Klasse: " + img1.getAttribute('alt') + " 2. Klasse: " + img2.getAttribute('alt');
			} else {
				return node.firstChild.nodeValue + " " ;
			}
		}

		//------------------------------------------------------------------------------
		// Sets variable "fussweg" as a closure.
		function fMproducts(node) {
			var value;
			if (node.childNodes[1] != null) {
				value = node.childNodes[1].getAttribute('alt') + " " ;
			} else {
				value = node.firstChild.nodeValue + " " ;
			}
			fussweg = value.match(/Fussweg/);
			return abbrevProduct(value);
		}

		//------------------------------------------------------------------------------
		function fMstopStation(node) {
			var value;
			if (node.childNodes[1] != null) {
				value = node.childNodes[1].firstChild.nodeValue;
			} else {
				value = node.firstChild.nodeValue;
			}
			return abbrevStation(value);
		}

		//------------------------------------------------------------------------------
		function abbrevStation(str) {
			return str
				.replace(/strasse/i, "str.")
				.replace(/platz/i, "pl.")
			;
		}

		//------------------------------------------------------------------------------
		function abbrevProduct(str) {
			return str
				.replace(/Tram/, "Trm")
			;
		}
	} // function initializeFunctionMap

	//------------------------------------------------------------------------------
	function appendRow(itiRow, squeezed, itinerary) {
			squeezed = preprocessSqueeze(squeezed);
			var newFinalRow = document.createElement('tr');
			var td = document.createElement('td');
			newFinalRow.appendChild(td);
			td.setAttribute("colspan", "9");
			// td.innerHTML =  "<table bgcolor='#FFFF88' id='iti" + itinerary + "'>" +squeezed + "</table>";
			td.innerHTML =  "<table id='iti" + itinerary + "'>" +squeezed + "</table>";

			itiRow.parentNode.appendChild(td);
	}

	//------------------------------------------------------------------------------
	function preprocessSqueeze(squeezed) {
		var squeezedLine = "";
//		GM_log("squeezed:"+ squeezed);
		var station1 = squeezed[1][3]; // FIXME: as soon as BugFirstLineEmpty fixed start with [0][3] instead of [1][3]
		
		var arr = station1.split(/, /);
		var city1 = arr[0];

		var sameCity = true;
		if(station1 != city1){ // there's a city before a station
			var i = 2; // FIXME: as soon as BugFirstLineEmpty fixed start with 1 instead of 2
			while(i < squeezed.length && sameCity){
				arr = squeezed[i][3].split(/, /);
				sameCity = arr[0] == city1;
				++i;
			}
			if(sameCity){ // all entries start with the same city
				// FIXME: as soon as BugFirstLineEmpty fixed start with 0 instead of 1
				for(i = 1; i < squeezed.length; ++i){
					arr = squeezed[i][3].split(/, /);
					squeezed[i][3] = arr[1];
				}
			}
		}

		for(var i = 0; i < squeezed.length; ++i){
			squeezedLine += "<tr><td>" + squeezed[i].join(" ") + "</td></tr>";
		}
		return squeezedLine;
	}

} // parseDetails


//---- functions ---------------------------------------------------------------

//------------------------------------------------------------------------------
function removeElementAtXpath(xpath){
	if(!(xpath instanceof Array)){
		xpath = new Array(xpath);
	}

	for(var i= 0; i < xpath.length; ++i){
		var nodes = getElementsAtXpath(xpath[i]);

		for(var j= 0; j < nodes.length; ++j){
			nodes[j].parentNode.removeChild(nodes[j]);
		}
	}
}

//------------------------------------------------------------------------------
function getElementsAtXpath(xpath){
	return getElementsAtXpathEl(xpath, document);
}

//------------------------------------------------------------------------------
function getElementsAtXpathEl(xpath, element){
	return coll2arr(
		document.evaluate(
			xpath,
			element,
			null,
			XPathResult.UNORDERED_NODE_ITERATOR_TYPE,
			null
		)
	);
}

//------------------------------------------------------------------------------
function coll2arr(coll) {
	var nodes = new Array();
	var i = 0;

	var thisNode;
	while ((thisNode = coll.iterateNext())) {
		nodes[i++] = thisNode;
	}
	return nodes;
}

//------------------------------------------------------------------------------
function removeAdStuff() {
	// don't reorder the xpaths!
	// TODO: figure out why it matters.
	removeElementAtXpath( [
		"/html/body/div[@class='posHead']",
		"/html/body/div[@class='posMenu']",
		"/html/body/div[@class='posFooter']",
		"//table[@class='hac_greybox']/tbody/tr/td[@class='hac_greybox_cell footer']",
		"//td[@class='hafas']/table[@class='separator-blue']",
		"//table[@cellspacing='0'][@cellpadding='0'][@border='0'][@width='100%']",
		"//td[@valign='top'][@style='width: 166px;']", // google ad
		"//div[@class='adv_group partner_area first']", // wetter
		"//div[@class='adv_group partner_area']", //
		"//iframe[@src='new_files/index.html']", // Smartbox
		"//td[@class='posInformationen']/table[@class='separator-blue'][last()]", // Partner
	] );
}

//------------------------------------------------------------------------------
// not used, here for demo purposes, or once I know how to invoke a download
// on click using client-side javascript only.
function replaceKalender() {
	var nodes = getElementsAtXpath('//a[string(.) = "Kalender"]');
	for (i = 0; i < nodes.length; ++i){
		var elmNewContent = document.createElement('a');
		elmNewContent.appendChild(document.createTextNode('Replaced!'));
		nodes[i].parentNode.replaceChild(elmNewContent, nodes[i]);
	}
}


//------------------------------------------------------------------------------
// http://4umi.com/web/javascript/select.php
//
function provideSelectionSourceCode() {

	var Utils = {
		NOT_SUPPORTED : {},
		DOM : {
			getElementWithId : function() {
				GM_log("DOM.getElementWithId");
				var func=function() { return Utils.NOT_SUPPORTED; }
				if(document.getElementById) {
					func=function(id) {
						return document.getElementById(id);
					}
				} else if(document.all) {
					func=function(id) {
						return document.all[id];
					}
				}
				GM_log("DOM.getElementWithId 2");
				return (this.getElementWithId=func)();
			}
		},
		Ranges : {
			create : function() {
				GM_log("Rangeѕ.create");
				var func=function() { return Utils.NOT_SUPPORTED};
				if(document.body && document.body.createTextRange) {
					func=function() { return document.body.createTextRange(); }
				} else if(document.createRange) {
					func=function() { return document.createRange(); }
				}
				GM_log("Rangeѕ.create 2");
				return (this.create=func)();
			},
			selectNode : function(node, originalRng) {
				GM_log("Rangeѕ.selectNode");
				var func=function() { return Utils.NOT_SUPPORTED; };
				var rng=this.create(), method="";
				if(rng.moveToElementText) method="moveToElementText";
				else if(rng.selectNode) method="selectNode";
				if(method) {
					func=function(node, rng) {
						rng=rng||Utils.Ranges.create();
						rng[method](node);
						return rng;
					}
				}
				GM_log("Rangeѕ.selectNode 2");
				return rng=null,(this.selectNode=func)(node, originalRng);
			}
		},
		Selection : {
			clear:function() {
				GM_log("Selection.clear ");
				var func=function() { return Utils.NOT_SUPPORTED};
				if(typeof document.selection!="undefined") {
					func=function() {
						if(document.selection && document.selection.empty) {
							return (Utils.Selection.clear=function() {
								if(document.selection)
									document.selection.empty();
							})();
						}
					}
				} else if(window.getSelection) {
					var sel=window.getSelection();
					if(sel.removeAllRanges) {
						func=function() {
							window.getSelection().removeAllRanges();
						}
					}
					sel=null;
				}
				GM_log("Selection.clear 2");
				return (this.clear=func)();
			},
			add : function(originalRng) {
				GM_log("Selection.originalRng");
				var func=function() { return Utils.NOT_SUPPORTED};
				var rng=Utils.Ranges.create();
				if(rng.select) {
					func=function(rng) {rng.select(); }
				} else if(window.getSelection) {
					var sel=window.getSelection();
					if(sel.addRange) {
						func=function(rng) {window.getSelection().addRange(rng); }
					}
					sel=null;
				}
				GM_log("Selection.originalRng 2");
				return (this.add=func)(originalRng);
			}
		}
	};

	addSelectionCode();

	function wrapInTimeOut(f, ms) {
		return function() {
			window.setTimeout(f, ms);
		}
	}

	function doSelection(id) {
		Utils.Selection.clear();
//		GM_log("selection cleared");
//		GM_log("should add element with id "+id);
//		GM_log("should add element "+document.getElementById(id));
		var rng = Utils.Ranges.create();
		var ele = document.getElementById( id )

		if(rng !== Utils.NOT_SUPPORTED && ele !== Utils.NOT_SUPPORTED) {
			Utils.Selection.add(
				Utils.Ranges.selectNode(
						document.getElementById(id)
				)
			);
//			GM_log("selection done");
		} else {
			GM_log("selection not supported: rng: " + rng == Utils.NOT_SUPPORTED + " ele "+ ele == Utils.NOT_SUPPORTED);
		}
	}

	function makeTextAndSelect(n) {
		var f = function() {
			parseDetails(n);
			doSelection("iti" + n);
		};
		return wrapInTimeOut(f, 500); 
		// don't lower this or find a better way to
		// find out when the reload has completed!
		// Can I add a callback?
	}

	//------------------------------------------------------------------------------
	function addSelectionCode(){
		var asbuttons = getElementsAtXpath('//a[starts-with(@id,"linkDtlOpenC")]');
		for(i = 0; i < asbuttons.length; ++i){

			// see http://www.quirksmode.org/js/events_order.html
			asbuttons[i].addEventListener("click", makeTextAndSelect(i),false); // true = capturing, false = bubbling

			// TODO: figure out a better way to scroll to the node without losing selection.
			// (avoid using timer)
		}
	}

}

